function data = generateToydata( varargin )
% Used to generate test manifolds data with d=1/2 and D=2/3.
%
% Returns nSet sets of nPoint data points each, where each set lies on a
% d-dimensional manifold in D-dimensional space, and the different
% manifolds on which the point sets lie are 'similar'. A good example of
% this is a series of concentric spheres, where points sampled from each
% sphere represent a single set. Alternatively, this func can be to simply
% generate commonly used manifolds such as the "swiss-roll" or "S-curve".
%
% If gtFlag=1 we can often obtain ground truth (gt) geodesic distances GD
% between points using the methods outlined in Sec 3.1 of ICML07. In
% addition to X, we sample a second larger set Xinf which is used to obtain
% the GD for each pair of points in X (using Dijkstra's). Xinf can also be
% used to obtain denoised versions of each point (if noise>0).
%
% The flag 'type' controls the type of data generated. See table below. If
% mul=T can generate multiple sets of this type (ow nSet defaults to 1). If
% gt=T can generate ground truth info (ow gtFlag defaults to 0). samp
% determines how points are sampled from manifolds. Possibilities include:
%  type   mul  gt    dims      samp          data-description
%  [-1]    T   T  (D=2/d=1)  UNIFORM      images of 1d gaussians
%  [0]     T      (D=2/d=1)  PATHS        sinusoidal
%  [1]     T      (D=2/d=1)  PATHS        circles
%  [1.1]   T      (D=2/d=1)  PATHS        circles [Q1]
%  [2]     T      (D=3/d=2)  PATHS        spheres
%  [3]     T   T  (D=2/d=1)  UNIFORM      circles
%  [3.1]   T   T  (D=2/d=1)  UNIFORM      circles [Q1]
%  [4]     T   T  (D=3/d=2)  UNIFORM      hemisphere
%  [4.1]   T   T  (D=3/d=2)  UNIFORM      sphere
%  [4.2]   T   T  (D=3/d=2)  UNIFORM      fishbowl
%  [4.3]   T   T  (D=3/d=2)  UNIFORM      hemisphere + hole
%  [5]         T  (D=3/d=2)  UNIFORM      S-curve
%  [6]         T  (D=3/d=2)  UNIFORM      Swiss-roll
%  [7]            (D=3/d=2)  STRUCTURED   S-curve
%  [8]            (D=3/d=2)  PATHS        Swiss-roll
%  [9]     T   T  (D=3/d=2)  UNIFORM      plane
%  [10]        T  (D=3/d=2)  UNIFORM      mobius strip
% Valid types include: [-1 0 1 1.1 2 3 3.1 4 4.1 4.2 4.3 5 6 7 8 9 10].
%
% USAGE
%   data = generateToydata( varargin )
%
% INPUT
%   varargin  - prm struct or name/value list w following fields:
%       'type'     - [REQ] see above
%       'nSet'     - [1]   number of point sets (often nSet must be 1)
%       'nPoint'   - [100] number of points per set
%       'noise'    - [0]   std of gaussian noise to add to data
%       'show'     - [0]   figure to use for display or 0 for none
%       'gtFlag'   - [0]   if 1 return ground truth (not always possible)
%       'gtnSamp'  - [1e3] num samples for the ground truth data Xinf
%
% OUTPUT
%   data      -  data struct
%       .X3        - [nSet x nPoint x D] data matrix
%       .X         - [D x (nSet*nPoint)] reshaped data matrix
%       .S3        - surfaces for visualization (see visualizeManifold)
%       .D         - [scalar] data dimension (high dim)
%       .d         - [scalar] manifold dimension (low dim)
%       .GD        - [nPoint x nPoint] geodesic dists (if nSet==1 & gtFlag)
%       .Xgt       - [D x (nSet*nPoint)] denoised data points (if gtFlag)
%
% EXAMPLE
%   data = generateToydata('type',5,'nSet',1,'nPoint',100,'show',1);
%   data = generateToydata('type',4,'nSet',3,'nPoint',250,'show',2);
%
% See also VISUALIZEMANIFOLD, COMPUTENEIGHBOR, DIJKSTRA
%
% LSML Toolbox     Version 1.1
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

ptd = getPrmDflt( varargin, {'type','REQ', 'nSet',1, 'nPoint',100, ...
  'noise',0, 'show',0, 'gtFlag',0, 'gtnSamp', 1000 });
type=ptd.type;  nSet=ptd.nSet;  nPoint=ptd.nPoint;  noise=ptd.noise;
show=ptd.show;  gtFlag=ptd.gtFlag;  gtnSamp=ptd.gtnSamp;

switch type
  case -1 %%% UNIFORM images of 1d gaussians
    D = 2; d = 1; gtValid = 1;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    for s=1:nSet
      if(s>1); sig=(rand*5+.1); else sig=1; end;
      x = randn( nPoint, 1 );
      y = exp(-x.^2 )*sig;
      X3(s,:,1)=x'; X3(s,:,2)=y';
      S3(s).X = linspace(-4,4,100);
      S3(s).Y = exp(-S3(s).X.^2 )*sig;
    end
    
  case 0 %%% PATHS sinusoidal (a+t,sin(a+t)+b)
    D = 2; d = 1; gtValid = 0;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    for s=1:nSet
      a=rand*10;  b=rand*10;  a1=a;
      for p=1:nPoint
        if( p>1 ); a = a + rand/2; end;
        X3(s,p,1:2) = [a sin(a)+b];
      end;
      S3(s).X = linspace(a1,a,100);
      S3(s).Y = sin(S3(s).X)+b;
    end;
    
  case 1 %%% PATHS circles
    D = 2; d = 1;  gtValid = 0;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    for s=1:nSet
      r=rand*10+.1;  theta=rand*10;  theta1=theta;
      for p=1:nPoint
        if( p>1 ); theta = theta + (.01+rand/r); end;
        X3(s,p,1) = r * cos( theta );
        X3(s,p,2) = r * sin( theta );
      end;
      t = linspace(theta1,theta,100);
      S3(s).X = r*cos(t);
      S3(s).Y = r*sin(t);
    end;
    
  case 1.1 %%% PATHS circles [Q1]
    D = 2; d = 1; gtValid = 0;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    minTheta=inf; maxTheta=-inf; rads=zeros(nSet,1);
    for s=1:nSet
      r=rand*10+5;  theta=rand*pi/2;  theta1=theta;
      for p=1:nPoint
        if( p>1 ); theta = theta + (.01+rand/r); end;
        theta=max(0,min(theta,pi/2));
        X3(s,p,1) = r * cos( theta );
        X3(s,p,2) = r * sin( theta );
      end;
      rads(s)=r;
      minTheta = min(minTheta,theta1);
      maxTheta = max(maxTheta,theta);
    end;
    t = linspace(minTheta-pi/10,maxTheta+pi/10,100);
    for s=1:nSet
      S3(s).X = ( rads(s)*cos(t) );
      S3(s).Y = ( rads(s)*sin(t) );
    end
    
  case 2 %%% PATHS spheres
    % http://en.wikipedia.org/wiki/Spherical_coordinates
    D = 3; d = 2; gtValid = 0;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    for s=1:nSet
      r = rand*10+1; theta=rand*10;  phi=rand*10;
      dir=randn(2,1);  dir=dir/norm(dir);
      for p=1:nPoint
        dir=dir+randn(2,1)/5;  dir=dir/norm(dir);  dx=rand/r;
        phi=phi+dir(1)*dx;  theta=theta+dir(2)*dx;
        X3(s,p,1) = r * sin( phi ) * cos( theta );
        X3(s,p,2) = r * sin( phi ) * sin( theta );
        X3(s,p,3) = r * cos( phi );
      end;
      [phis,thetas]=ndgrid( 0:.1:(pi+.1), 0:.1:(2*pi+.1) );
      S3(s).X = r .* sin( phis ) .* cos( thetas );
      S3(s).Y = r .* sin( phis ) .* sin( thetas );
      S3(s).Z = r .* cos( phis );
    end;
    
  case 3 %%% UNIFORM circles
    D = 2; d = 1; gtValid = 1;
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    X3 = randn( nSet, nPoint, D );
    X3 = X3 ./ repmat(sqrt(sum(X3.^2,3)),[1 1 D]);
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      X3(s,:,:) = X3(s,:,:)*r;
      thetas = 0:.05:(2*pi+.1);
      S3(s).X = r .* sin( thetas );
      S3(s).Y = r .* cos( thetas );
    end;
    
  case 3.1 %%% UNIFORM circles [Q1]
    D = 2; d = 1; gtValid = 1;
    X3 = randn( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[]),1,nSet);
    X3 = abs(X3);
    X3 = X3 ./ repmat(sqrt(sum(X3.^2,3)),[1 1 D]);
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      X3(s,:,:) = X3(s,:,:)*r;
      thetas = 0:.1:(pi/2+.05);
      S3(s).X = r .* sin( thetas );
      S3(s).Y = r .* cos( thetas );
    end;
    
  case 4 %%% UNIFORM hemisphere
    D = 3; d = 2; gtValid = 1;
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    X3 = randn( nSet, nPoint, D );
    X3(:,:,3) = abs(X3(:,:,3));
    X3 = X3 ./ repmat(sqrt(sum(X3.^2,3)),[1 1 D]);
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      X3(s,:,:) = X3(s,:,:)*r;
      [phis,thetas]=ndgrid( 0:.1:(pi/2+.1), -pi:.1:(pi+.1) );
      S3(s).X = r .* sin( phis ) .* cos( thetas );
      S3(s).Y = r .* sin( phis ) .* sin( thetas );
      S3(s).Z = r .* cos( phis );
    end;
    
  case 4.1 %%% UNIFORM sphere
    D = 3; d = 2; gtValid = 1;
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    X3 = randn( nSet, nPoint, D );
    X3 = X3 ./ repmat(sqrt(sum(X3.^2,3)),[1 1 D]);
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      X3(s,:,:) = X3(s,:,:)*r;
      [phis,thetas]=ndgrid( 0:.1:(pi+.1), -pi:.1:(pi+.1) );
      S3(s).X = r .* sin( phis ) .* cos( thetas );
      S3(s).Y = r .* sin( phis ) .* sin( thetas );
      S3(s).Z = r .* cos( phis );
    end;
    
  case 4.2 %%% UNIFORM fishbowl
    D = 3; d = 2; gtValid = 1;
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    X3 = zeros( nSet, nPoint, D );
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      p=0;
      while( p < nPoint )
        v = randn(3,1); v=v/norm(v);
        if( v(3)>-.2 ); p=p+1; X3(s,p,:) = v; end;
      end;
      [phis,thetas]=ndgrid( 0:.1:(pi*.6+.1), 0:.1:(2*pi+.1) );
      S3(s).X = r .* sin( phis ) .* cos( thetas );
      S3(s).Y = r .* sin( phis ) .* sin( thetas );
      S3(s).Z = r .* cos( phis );
    end;
    
  case 4.3 %%% UNIFORM hemisphere + hole
    D = 3; d = 2; gtValid = 1;
    X3 = zeros( nSet, nPoint, D );
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    for s=1:nSet
      if(s>1); r = (rand+.1); else r=1; end;
      p=0;
      while( p < nPoint )
        v = randn(3,1); v=v/norm(v); v(3) = abs(v(3));
        if( v(3)<0.9 ); p=p+1; X3(s,p,:) = v; end;
      end;
      [phis,thetas]=ndgrid( 0:.1:(pi*.6+.1), 0:.1:(2*pi+.1) );
      S3(s).X = r .* sin( phis ) .* cos( thetas );
      S3(s).Y = r .* sin( phis ) .* sin( thetas );
      S3(s).Z = r .* abs( cos( phis ) );
    end;
    
  case 5 %%% UNIFORM S-curve
    D = 3; d = 2; nSet = 1; gtValid = 1;
    angle = 3*pi*rand(1,nPoint);
    highAngle = ( angle>=1.5*pi );
    angle = angle - pi - pi/2*highAngle;
    X3(1,:,:)=[cos(angle); 2*rand(1,nPoint); sin(angle)+2*highAngle ];
    X3 = permute( X3, [ 1 3 2 ] );
    tt = (-1:0.1:0.5)*pi; uu = tt(end:-1:1); hh = (0:0.05:1)*2;
    S3(1).X = [cos(tt) -cos(uu)]'*ones(size(hh));
    S3(1).Y = ones(size([tt uu]))'*hh;
    S3(1).Z = [sin(tt) 2-sin(uu)]'*ones(size(hh));
    
  case 6 %%% UNIFORM swiss-roll
    D = 3; d = 2; nSet = 1; gtValid = 1;
    tt = (3*pi/2)*(1+2*rand(1,nPoint));
    height = 20*rand(1,nPoint);
    X = [tt.*cos(tt); height; tt.*sin(tt)];
    X3 = permute(X,[3 2 1]);
    tt = (3*pi/2)*(1+2*(0:0.02:1)); hh = (0:0.125:1)*20;
    S3(1).X = (tt.*cos(tt))'*ones(size(hh));
    S3(1).Y = ones(size(tt))'*hh;
    S3(1).Z = (tt.*sin(tt))'*ones(size(hh));
    
  case 7 %%% STRUCTURED S-curve
    D = 3; d = 2; nSet = 1; gtValid = 0;
    tt = linspace(-1,.5,20)*pi; hh=0:1:2;
    uu = tt(end:-1:1);
    a=[cos(tt) -cos(uu)]'*ones(size(hh));
    b=ones(size([tt uu]))'*hh;
    c=[sin(tt) 2-sin(uu)]'*ones(size(hh));
    X31 = cat(3,a(:)',cat(3,b(:)',c(:)'));
    tt = tt(1:4:end);  hh = linspace(0,2,8);
    uu = tt(end:-1:1);
    a=[cos(tt) -cos(uu)]'*ones(size(hh));
    b=ones(size([tt uu]))'*hh;
    c=[sin(tt) 2-sin(uu)]'*ones(size(hh));
    X32 = cat(3,a(:)',cat(3,b(:)',c(:)'));
    X3 = [X31 X32];
    tt = (-1:0.01:0.5)*pi; uu = tt(end:-1:1); hh = (0:0.01:1)*2;
    S3(1).X = [cos(tt) -cos(uu)]'*ones(size(hh));
    S3(1).Y = ones(size([tt uu]))'*hh;
    S3(1).Z = [sin(tt) 2-sin(uu)]'*ones(size(hh));
    
  case 8 %%% PATHS swiss-roll
    D=3; d = 2; gtValid = 0;
    X3 = zeros( 1, nSet*nPoint, D );
    bnds = [3*pi/2 3*pi; 0 20];  cnt=1;
    for s=1:nSet
      v = rand(2,1).*bnds(:,2) + bnds(:,1);
      dir = randn(2,1);  dir=dir/norm(dir);
      for p=1:nPoint
        dir=dir+randn(2,1)/5;  dir=dir/norm(dir);
        dx=rand/20;
        v(1)=v(1)+dir(1)*dx; v(2)=v(2)+10*dir(2)*dx;
        for i=1:2
          v(i) = max(bnds(i,1),min(v(i),bnds(i,1)+bnds(i,2)));
          if(v(i)==bnds(i,1)); dir(i)=abs(dir(i)); end;
          if(v(i)==bnds(i,1)+bnds(i,2)); dir(i)=-abs(dir(i)); end;
        end;
        X3(1,cnt,1:3) = [v(1)*cos(v(1)); v(2); v(1)*sin(v(1))];
        cnt=cnt+1;
      end;
    end;
    tt = (3*pi/2)*(1+2*(0:0.02:1)); hh = (0:0.125:1)*20;
    S3(1).X = (tt.*cos(tt))'*ones(size(hh));
    S3(1).Y = ones(size(tt))'*hh;
    S3(1).Z = (tt.*sin(tt))'*ones(size(hh));
    
  case 9 %%% UNIFORM plane
    D = 3; d = 2; gtValid = 1;
    S3 = repmat(struct('X',[],'Y',[],'Z',[]),1,nSet);
    X3 = zeros( nSet, nPoint, D );
    for s=1:nSet
      X3(s,:,1) = rand( 1, nPoint );
      X3(s,:,2) = rand( 1, nPoint );
      if(s>1); z = rand; else z=1; end;
      X3(s,:,3) = z;
      S3(s).X = (0:0.1:1)'*ones(1,11);
      S3(s).Y = ones(1,11)'*(0:0.1:1);
      S3(s).Z = X3(s,1,3)*ones(11,11);
    end;
    
  case 10 %%% UNIFORM mobius strip
    D = 3; d = 2; nSet = 1; gtValid = 1;
    X3 = zeros( nSet, nPoint, D );
    mw=0.05; br=0.5;
    u = (rand(1,nPoint)-0.5)*2*pi; v = (rand(1,nPoint)-0.5)*2*pi;
    X3(1,:,1) = br*ones(size(v))+br*sin(v)+mw*u.*sin(v).*cos(v/2);
    X3(1,:,2) = mw*u.*sin(v/2);
    X3(1,:,3) = br*cos(v)-mw*u.*cos(v).*cos(v/2);
    [u,v]=meshgrid(-pi:0.1:pi+0.1,-pi:0.1:pi+0.1);
    S3(1).X = br*ones(size(v))+br*sin(v)+mw*u.*sin(v).*cos(v/2);
    S3(1).Y = mw*u.*sin(v/2);
    S3(1).Z = br*cos(v)-mw*u.*cos(v).*cos(v/2);
    
  otherwise %%% unknown
    error('invalid type');
end;


%%% Add noise if requested:
if( noise~=0 )
  X3 = X3 + randn(size(X3)) * noise;
end;

%%% store results
X = reshape(permute(X3,[2 1 3]),[],size(X3,3))';
data = struct( 'X3',X3, 'X',X, 'S3',S3, 'D',D, 'd',d );

%%% ground truth
if(gtFlag && ( ~gtValid || nSet~=1 ))
  gtFlag = false;
  warning('gt invalid (possibly due to nSet>1)'); %#ok<WNTAG>
end
if( gtFlag )
  % generate many points by repeated self calls
  nSampMax = min( 100, gtnSamp );
  ptdGt=ptd; ptdGt.nPoint=nSampMax;  ptdGt.show=0;
  ptdGt.gtFlag=0; ptdGt.noise=0;
  Xinf=generateToydata(ptdGt);  Xinf=Xinf.X;
  X3inf=permute( reshape(Xinf',nSampMax,nSet,D), [2 1 3] );
  neigh = computeNeighbor(X3inf,'k',1);
  gtEps=neigh.scale*5; iter=0;
  while( size(Xinf,2)<gtnSamp && iter<1000 )
    Xi=generateToydata(ptdGt); Xi=Xi.X;
    [ind,indi] = find( pdist2(Xinf',Xi')<gtEps^2 );
    indi=unique(indi);  Xi(:,indi)=[];
    Xinf = [Xinf Xi]; %#ok<AGROW>
    if( length(indi)>(.95*nSampMax) );
      gtEps = gtEps*.9;
    end;
    iter=iter+1;
  end
  Xinf = Xinf(:,1:min(size(Xinf,2),gtnSamp));
  
  % get 'denoised' points (closest non-noisy point on manifold)
  [ dis, ind ] = min( pdist2(Xinf',data.X'), [], 1 );
  data.Xgt = Xinf(:,ind);
  
  % compute geodesic distances using Xinf
  X3inf=permute( reshape(Xinf',gtnSamp,nSet,D), [2 1 3] );
  if(0); visualizeManifold( X3inf, show+1 ); end;
  neigh = computeNeighbor( X3inf,'k',50 );
  GD = zeros( length( ind ) );
  for i = 1 : length(ind)
    dist = dijkstra( neigh.Dmat, ind(i) )';
    GD( i, : ) = dist( ind );
  end
  GD = 0.5*(GD + GD');
  data.GD = GD;
end

%%% optionally display
if( show )
  visualizeManifold( X3, show, 'S3',S3 );
end
