function [H,ftrs] = lsmlComputeH( X, manifold, orthFlag )
% Compute tangent vectors at points X.
%
% This function computes the vectors tangent to the manifold for every data
% point (as well as every set of RBF features). The tangent vectors are
% simply computed with the linear model given before eqn. (5) in ICML07:
%  H_theta(x^{ij}) = [ \Theta^1 f^{ij} ... \Theta^D f^{ij} ]^\top
% The theta's are given in manifold.th and the f's are computed using
% rbfComputeFtrs and the basis in manifold.rbfBasis.
%
% USAGE
%   [H,ftrs] = lsmlComputeH( X, manifold, [orthFlag] )
%
% INPUT
%   X or X3   - data matrix: [D x n] or [nSet x nPoint x D]
%   manifold  - struct returned by lsmlOptimizeTh
%   orthFlag  - [false] if true orthonormalizes H
%
% OUTPUT
%   H         - [D x d x n] d tangent vectors at each point
%   ftrs      - [f x n] features for each point
%
% EXAMPLE
%
% See also LSMLOPTIMIZETH, VISUALIZEMANIFOLD
%
% LSML Toolbox     Version 1.1
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

% get inputs
if( nargin<3 || isempty(orthFlag)); orthFlag=0; end
n=size(X,2); d=manifold.d; D=manifold.D;
if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],D)'; end

% compute features
if(manifold.rbfFlag)
  ftrs=rbfComputeFtrs(X',manifold.rbfBasis)';
else
  ftrs=X;
end

% compute H
H = zeros(D,d,n);
for k=1:D; H(k,:,:)=manifold.th(:,:,k)*ftrs; end
%H = permute(multiprod(manifold.th,ftrs),[ 3 1 2 ]); %slow

% optionally orthonormalize
if( orthFlag )
  for i=1:n, H(:,:,i)=orth1(H(:,:,i),d); end
end

function Q = orth1( A, d )
% same as orth but ~3x faster (doesn't use svd)
[Q,R]=qr(A); Q=Q(:,1:d); %#ok<NASGU>
