function lsmlDemo( demoNumber )
% Demo for several abilities of LSML.
%
% Run with different integer values of demoNumber to run different demos.
%
% The demos are as follows:
%  1: Generate several data sets and display them.
%  2: Compute nearest neighbors and display them.
%  3: Learn tangents on a manifold.
%  4: Compute geodesic distances on a manifold.
%  5: Compute low-dim embedding of a manifold.
%  6: Denoise points on a manifold.
%
% USAGE
%  lsmlDemo( demoNumber )
%
% INPUTS
%  demoNumber - [1] value between 1 and 6
%
% OUTPUTS
%
% EXAMPLE
%  lsmlDemo(1)
%
% See also
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

if(nargin<1), demoNumber=1; end; c
if(demoNumber<1 || demoNumber>6), error('Invalid demo number.'); end
disp('Demos of various functions in LSML toolbox.');
disp('Steps marked w ''*'' or ''**'' may take time, please be patient.');
disp('Note: when entering input, focus must be on figure.');
disp('------------------------------------------------------------------');
fprintf('BEGIN DEMO %i: ',demoNumber);
eval(['demo' int2str(demoNumber) '()']);
disp(['END OF DEMO ' int2str(demoNumber) '.']);
disp('------------------------------------------------------------------');
disp('Press (n) to continue to next demo or (r) to repeat demo.');
drawnow; w=waitforbuttonpress; if(w==0), return; end
if(w==1 && get(gcf,'CurrentCharacter')=='n' && demoNumber<6),
  lsmlDemo(demoNumber+1);
elseif(w==1 && get(gcf,'CurrentCharacter')=='r'),
  lsmlDemo(demoNumber);
end
end

function demo1() %#ok<DEFNU>
disp('generateToydata()');
disp('Generate and show 1000 points sampled from ''S-curve''.');
generateToydata('type',5,'nPoint',1000,'show',1);
spin3d(); demoPause();

disp('Generate 1000 NOISY points sampled from ''swiss-roll''.');
generateToydata('type',6,'nPoint',1000,'noise',1,'show',1);
spin3d(); demoPause();

disp('Generate points sampled from 2 concentric hemispheres.');
generateToydata('type',4,'nSet',2,'nPoint',250,'show',1);
spin3d(); demoPause();

disp('Generate points along 25 ''paths'' sampled from ''swiss-roll''.');
generateToydata('type',8,'nSet',25,'nPoint',50,'show',1);
spin3d();
end

function demo2() %#ok<DEFNU>
disp('computeNeighbor() and visualizeManifold()');
disp('First we generate some toy data');
data = generateToydata('type',10,'nSet',1,'nPoint',200,'show',1);
spin3d(); demoPause();

disp('Compute k-nearest neighbors using k=2');
neigh = computeNeighbor( data.X3,'k',5,'show',1);
demoPause();

disp('Redisplay neighbors, overlayed on manifold representation.');
visualizeManifold(data.X3,1,'S3',data.S3,'N',neigh.N);
demoPause();

disp(['Median dist to nearest neighbor is ' num2str(neigh.scale)]);
disp('Compute nearest neighbors using maxDist=5*median dist');
computeNeighbor( data.X3,'maxDist',neigh.scale*5,'show',1);
demoPause();

disp('Compute NN using minimum k so that graph is fully connected.');
neigh = computeNeighbor( data.X3,'k',1,'forceConn',1,'show',1);
fprintf('k=%i resulted in a fully connected graph.\n', neigh.k);
end

function demo3() %#ok<DEFNU>
disp('lsmlOptimizeTh() and lsmlComputeH()');
disp('First we generate some toy data and get neighbor connections.');
data = generateToydata('type',10,'nSet',1,'nPoint',100,'show',0);
neigh = computeNeighbor( data.X3,'k',3,'forceConn',2,'show',1);
demoPause();

disp('*Learn rbfs and thetas, running 5 restarts.');
pTh=struct('d',data.d,'rbfK',5,'nRestart',5,'nSamples',250,'show',1);
M=lsmlOptimizeTh(data.X3,neigh.Nmat,pTh);
demoPause();

disp('Compute and display tangent vectors at each point.');
H = lsmlComputeH( data.X3, M );
visualizeManifold(data.X3, 1, 'VF',H, 'S3',data.S3 );
spin3d(); demoPause();

disp('Generate new data points for testing, recompute tangents.');
data = generateToydata('type',10,'nSet',1,'nPoint',20,'show',0);
H = lsmlComputeH( data.X3, M );
visualizeManifold(data.X3, 1, 'VF',H, 'S3',data.S3 );
spin3d();
end

function demo4() %#ok<DEFNU>
disp('lsmlGeodesic()');
disp('*First generate data and learn the thetas.');
data = generateToydata('type',4,'nPoint',25,'gtFlag',1,'show',0);
neigh = computeNeighbor( data.X3,'maxDist',1,'forceConn',5,'show',0);
pTh = struct('d',2,'rbfK',10,'nRestart',10,'nSamples',1000,'show',1);
M=lsmlOptimizeTh(data.X3,neigh.Nmat,pTh); n=size(data.X,2);
visualizeManifold(data.X3,1,'VF',lsmlComputeH(data.X3,M),'S3',data.S3);
spin3d(); demoPause();

disp('Now compute/display snakes connecting random points.');
for i=1:20
  inds=randint2(1,2,[1 n]); X2=data.X(:,[inds(1) inds(2)]);
  [gd chi]=lsmlGeodesic(X2,M,'epsSnk',neigh.scale);
  visualizeManifold(zeros(3,0),1,'Snake',chi,'S3',data.S3); pause(.1);
end
demoPause();

disp('*Now compute all geodesics and measure error against gt.')
GD=lsmlGeodesic(data.X,M,'Nmat',neigh.Nmat);
fprintf('Resulting error of geodesics is %.2f%%\n',distErr(GD,data.GD));
demoPause();

disp('Now compute all geodesics using Dijkstra''s and measure error.')
GD=dijkstra( neigh.Dmat, 1:n );
fprintf('Resulting error of geodesics is %.2f%%\n',distErr(GD,data.GD));
end

function demo5() %#ok<DEFNU>
disp('lsmlEmbed()');
disp('*First generate sparse data and learn the thetas.');
data = generateToydata('type',5,'nPoint',75,'gtFlag',1);
neigh = computeNeighbor( data.X3,'maxDist',1.1,'forceConn',1,'show',0);
pTh = struct('d',2,'rbfK',10,'nRestart',10,'nSamples',2500,'show',1);
M = lsmlOptimizeTh(data.X3,neigh.Nmat,pTh);
visualizeManifold(data.X3,1,'VF',lsmlComputeH(data.X3,M),'S3',data.S3);
spin3d(); demoPause();

disp('**Get landmark points and compute embedding.');
[Y L]=lsmlEmbed(data.X,M,'d',2,'m',inf); D=pdist2(Y',Y','euclidean');
fprintf('Resulting error of embedding is %.2f%%\n',distErr(D,data.GD));
visualizeManifold(data.X3,1,'S3',data.S3,'Y',Y);
demoPause();

disp('**Generate test data and perform out of sample embedding.');
dataTest = generateToydata('type',5,'nPoint',20,'gtFlag',1);
Y=lsmlEmbed(dataTest.X,M,'d',2,'landmarks',L); D=pdist2(Y',Y','euclidean');
fprintf('Resulting error of embedding is %.2f%%\n',distErr(D,dataTest.GD));
visualizeManifold(dataTest.X3,1,'S3',dataTest.S3,'Y',Y);
demoPause();

disp('Finally, for comparison, embed same data using Isomap.');
[GD Y] = lsmlIsomap( neigh, 2 ); D=pdist2(Y',Y','euclidean');
fprintf('Resulting error of embedding is %.2f%%\n',distErr(D,data.GD));
visualizeManifold(data.X3,1,'S3',data.S3,'Y',Y);
end

function demo6() %#ok<DEFNU>
disp('lsmlDenoise()');
disp('First generate NOISY data (and also include ground truth).');
data=generateToydata('type',3,'nPoint',100,'gtFlag',1,'noise',.1,'show',1);
demoPause();

disp('*Next, compute thetas using the noisy data.');
neigh = computeNeighbor( data.X3,'maxDist',.5,'show',0);
pTh = struct('d',data.d,'rbfK',5,'nRestart',10,'nSamples',2500,'show',1);
M = lsmlOptimizeTh(data.X3,neigh.Nmat,pTh);
visualizeManifold(data.X3,1,'VF',lsmlComputeH(data.X3,M),'S3',data.S3);
demoPause();

disp('Now denoise manifold using technique from ICML07.');
pDenoise = struct('lambdaN',5, 'errDel',1e-4, 'show',1,'aveFlag',0);
Chi3 = lsmlDenoise(data.X3, neigh.Nmat, M, pDenoise);
visualizeManifold( data.X3, 1, 'S3',data.S3, 'Chi3',Chi3 );
err=sum(sum((data.Xgt-squeeze(Chi3)').^2));
fprintf('Resulting squared error is %.3f\n',err);
demoPause();

disp('Now denoise manifold using simple averaging filter.');
pDenoise.aveFlag=1; Chi3=lsmlDenoise(data.X3, neigh.Nmat, M, pDenoise);
visualizeManifold( data.X3, 1, 'S3',data.S3, 'Chi3',Chi3 );
err=sum(sum((data.Xgt-squeeze(Chi3)').^2));
fprintf('Resulting squared error is %.3f\n',err);
end

function err = distErr( D, Dgt )
err = mean(abs(D(Dgt>1e-10)./Dgt(Dgt>1e-10)-1))*100;
end

function spin3d(speed)
axis vis3d; if(nargin<1), speed=1; end
nRot=1; nDeg=ceil(5*speed); n=round(360*nRot/nDeg);
for i=1:n, camorbit(nDeg,0,'camera'); drawnow; end
end

function demoPause()
disp('---press any key to continue---');
drawnow; waitforbuttonpress;
end
