function [Chi3, Chi] = lsmlDenoise( X, Nmat, manifold, varargin )
% Denoise manifold points given learned manifold.
%
% This function simply removes the noise from points on a noisy manifold
% (that has been learned from the original noisy points). Denoising can be
% done in one of two ways: (1) a local averaging window (if aveFlag==1) and
% (2) LSML denoising as explained in lsmlOptimizeChi (ow).
%
% USAGE
%   [Chi3,Chi] = lsmlDenoise( X, manifold, prm )
%
% INPUT
%   X or X3     - data matrix: [D x n] or [nSet x nPoint x D]
%   Nmat        - [n x n] neighborhood matrix [see computeNeighbor]
%   manifold    - struct returned by lsmlOptimizeTh
%   varargin    - prm struct or name/value list w following fields:
%       .aveFlag   - [0] replace each pt by average of neighbors
%       .lambdaN   - [0] stickiness to orig pnts
%       .errDel    - [see lsmlOptimizeChi]
%       .nItr      - [see lsmlOptimizeChi]
%       .show      - [see lsmlOptimizeChi]
%
% OUTPUT
%   Chi3       - [nSet x nPoint x D] denoised data
%   Chi        - [D x n] denoised data
%
% EXAMPLE
%
% See also LSMLOPTIMIZETH LSMLOPTIMIZECHI COMPUTENEIGHBOR
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

%%% get inputs
prm = getPrmDflt( varargin, {'lambdaN',0 'aveFlag',0} );
if(ndims(X)==3);
  [nSet,nPoint,D] = size(X);  n=nSet*nPoint;
  X=reshape(permute(X,[2 1 3]),[],size(X,3))';
else
  [D,n]=size(X);  nSet=1;  nPoint=n;
end
N=cell(1,n);  for i=1:n; N{i}=find(Nmat(i,:));  end

if( prm.aveFlag )
  % replace each point by an average of its neighbors
  Chi=X;  for i=1:n; Chi(:,i)=mean(X(:,N{i}),2);  end
else
  % perform gradient descent (reshape Chi to Chi3)
  prm.N=N;  prm.errCase=0; prm.errDisp=1;
  Chi = lsmlOptimizeChi( X, manifold, prm );
end
Chi3 = permute( reshape(Chi',nPoint,nSet,D), [2 1 3] );
