function [Y L] = lsmlEmbed( X, manifold, varargin )
% Embed D-dim data into d-dim space while trying to preserve geodesics.
%
% Note that embedding can only be performed accurately for manifolds that
% are isometric to Euclidean space. The set of manifolds that satisfy these
% properties are extremely limited.
%
% The embedding relies on the use of landmark points, as in the paper:
%  Sparse multidimensional scaling using landmark points
%  Vin de Silva & Joshua B. Tenenbaum.
%  http://pages.pomona.edu/~vds04747/public/publications.html
% First, m landmarks are randomly selected and the geodesics are computed
% between every pair. For every remaining point i geodesics are computed
% from i to each landmark. Finally, the embedding of these points is
% computed using "Distance-based triangulation" (see Eqn. 3 of Silva04).
%
% The primary difference between Silva04 and lsmlEmbed is that geodesics
% are computed using gradient descent on snakes (lsmlGeodesic) instead of
% by using Dijkstra's. The accuracy of the resulting geodesics is much
% higher, especially if there are few points.
%
% USAGE
%   [Y landmarks] = lsmlEmbed( X, manifold, varargin )
%
% INPUT
%   X or X3   - data matrix: [D x n] or [nSet x nPoint x D]
%   manifold  - learned tangent space [see lsmlOptimizeTh]
%   varargin  - prm struct or name/value list w following fields:
%     [parameters for lsmlGeodesic]
%       'errDel'    - [see lsmlGeodesic]
%       'nItr'      - [see lsmlGeodesic]
%       'epsSnk'    - [see lsmlGeodesic]
%     [parameters for embedding]
%       'd'         - [] embedding dimension
%       'm'         - [50] number of landmarks to use
%       'landmarks' - landmark info (from previous embedding)
%
% OUTPUT
%   Y         - [d x n] coordinates of embedded points
%   landmarks - landmark struct
%      'm'          - number of landmarks
%      'X'          - [D x m] landmark points
%      'Y'          - [d x m] coordinates of embedded landmarks
%      'GD'         - geodesic distances between landmarks
%      'chis'       - chis corresponding to snakes between landmarks
%      'epsSnk'     - [see lsmlGeodesic]
%
% EXAMPLE
%
% See also CMDSCALE LSMLGEODESIC
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

% get parameters
[nItr errDel epsSnk d m L] = getPrmDflt( varargin, ...
  {'nItr',[] 'errDel',[] 'epsSnk',[] 'd',0 'm',0 'landmarks',[]});
pGd=struct('nItr',nItr, 'errDel',errDel, 'epsSnk',epsSnk );
if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],size(X,3))'; end
n=size(X,2);

% get landmarks structure to perform this/future embeddings
if( isempty(L) )
  assert(m>0); if(m<n), Xl=X(:,randsample(n,m)); else m=n; Xl=X; end
  if(isempty(epsSnk)), N=computeNeighbor(Xl,'k',1); epsSnk=N.scale; end
  [GD,chis] = lsmlGeodesic( Xl, manifold, pGd ); Y=cmdscale(GD)';
  L=struct('m',m, 'X',Xl, 'Y',Y, 'GD',GD, 'chis',{chis}, 'epsSnk',epsSnk);
  if(m==n), if(d), Y=Y(1:d,:); end; return; end
end

% compute gd from each point i to all m landmark points
if(isempty(pGd.epsSnk)), pGd.epsSnk=L.epsSnk; end
m=L.m; Xi=[L.X zeros(manifold.D,1)]; GD=zeros(m,n);
pGd.GD=sparse(m+1,m+1); pGd.GD(1:m,1:m)=L.GD;
pGd.chis=cell(m+1,m+1); pGd.chis(1:m,1:m)=L.chis;
ticid = ticStatus('embedding points');
for i=1:n
  [v,id]=min(pdist2(X(:,i)',L.X'));
  if(v==0), GD(:,i)=L.GD(:,id); continue; end
  for j=1:m+1; pGd.chis{m+1,j}=[]; pGd.chis{j,m+1}=[]; end
  pGd.GD(:,m+1)=0; pGd.GD(m+1,:)=0; Xi(:,m+1)=X(:,i);
  GDi=lsmlGeodesic(Xi, manifold, pGd); GD(:,i)=GDi(1:m,m+1);
  tocStatus(ticid,i/n);
end

% finaly embed all points
Z=-pinv(L.Y')/2; z=mean(L.GD.^2,2);
Y = Z*(GD.*GD-repmat(z,1,n));
if(d), Y=Y(1:d,:); end

end
