function [GD chis] = lsmlGeodesic( X, manifold, varargin )
% Compute the geodesic distance between points on the manifold.
%
% This function is actually a wrapper for two functions:
%  1) lsmlGeodesic1( X1, X2, manifold, varargin )
%  2) lsmlGeodesicAll( X, manifold, varargin )
% If size(X,2)==2 (ie X=[X1; X2]) the first usage is called.
%
% 1) lsmlGeodesic1(...): This function computes geodesic distance (gd)
% between a PAIR of points. The snake is initialized by connecting the two
% points and linearly interpolating points in between according to epsSnk.
% The snake is then OPTIMIZED (procedure described below). This operation
% is best suited for points that are relatively nearby.
%
% 2) lsmlGeodesicAll(...): This function computes geodesic distance between
% ALL PAIRS of points on the manifold. First, for all nearby points
% lsmlGeodesic1(...) is used to compute the snake between them. Next, for
% any three points i1-i12-i2 such that the geodesics between i1-i12 and
% i12-i2 are computed, the gd between i1-i2 is not, and the shortest path
% between i1 and i2 is i1-i12-i2, a snake is initialized between i1-i2 by
% merging the snakes from i1-i12 and i12-i2. This snake is then OPTIMIZED
% (procedure described below). This continues until all GDs are computed.
%
% The OPTIMIZATION for a single snake, after it has been initialized, is
% performed using the procedure described in Equations 10/11 of ICML07.
% Specifically, lsmlOptimizeChi() is used to optimize the snake so it has
% minimum length while remaining on the manifold. This snake is first
% optimized to be on the manifold (projection), then optimized to have
% minimum length while remaining on manifold.
%
% USAGE
%   [GD err chis] = lsmlGeodesic( X, manifold, prm )
%
% INPUT
%   X or X3     - data matrix: [D x n] or [nSet x nPoint x D]
%   manifold    - learned tangent space [see lsmlOptimizeTh]
%   varargin    - prm struct or name/value list w following fields:
%       'errDel'    - [see lsmlOptimizeChi]
%       'nItr'      - [see lsmlOptimizeChi]
%       'epsSnk'    - ['REQ'] step for points along snake (use neigh.scale)
%       'Nmat'      - [] neighbors for each point (should be connected)
%       'GD'        - [] include prev. computed geodesics (also need chis)
%       'chis'      - [] include prev. computed snakes (also need GD)
%
% OUTPUT
%   GD          - [nPoint x nPoint] geodesic distance matrix
%   chis        - [nPoint x nPoint] cell of snakes
%
% EXAMPLE
%
% See also LSMLOPTIMIZETH LSMLOPTIMIZECHI COMPUTENEIGHBOR GENERATETOYDATA
%
% LSML Toolbox     Version 1.1
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],size(X,3))'; end
if(size(X,2)==2 )
  prm = getPrmDflt(varargin,{'nItr',[] 'errDel',[] 'epsSnk','REQ'});
  [GD chi]=lsmlGeodesic1( X(:,1),X(:,2),manifold,prm,prm.epsSnk );
  chis={chi};
else
  [GD chis]=lsmlGeodesicAll(X,manifold,varargin{:});
end
end

function [GD chis] = lsmlGeodesicAll( X, manifold, varargin )
%%% get parameters
[nItr errDel epsSnk Nmat GD chis] = getPrmDflt( varargin, ...
  {'nItr',[] 'errDel',[] 'epsSnk',[] 'Nmat',[] 'GD',[] 'chis',[] });
prmSnk=struct('nItr',nItr, 'errDel',errDel); n=size(X,2);

%%% retrieve or create neighborhood matrix
if(isempty(Nmat)||isempty(epsSnk)), neigh=computeNeighbor(X,'k',1); end;
if(isempty(epsSnk)), epsSnk=neigh.scale; end
if(isempty(Nmat))
  prmN = struct('maxDist',neigh.scale,'forceConn',1,'show',0);
  neigh=computeNeighbor(X,prmN); Nmat=neigh.Nmat; clear neigh;
end

%%% initialize
if(isempty(GD)), done=eye(n); GD=sparse(n,n); chis=cell(n,n); else
  for i=1:n*n, assert(isempty(chis{i})==GD(i)==0); end
  done=GD>0|eye(n); assert(all(all(done==done')));
end

%%% compute GD between all neighboring pairs
for i1 = 1:n
  for i2=find( Nmat(i1,:) & ~done(i1,:));
    [gd chi]=lsmlGeodesic1(X(:,i1),X(:,i2),manifold,prmSnk,epsSnk);
    GD(i1,i2)=gd; GD(i2,i1)=gd; done(i1,i2)=1; done(i2,i1)=1;
    chis{i1,i2}=chi; chis{i2,i1}=fliplr(chi);
  end
end

%%% compute GD between all remaing pairs
ticid = ticStatus('geodesic distance'); isDiff=1;
while(isDiff), isDiff=0;
  for i1 = 1:n
    % get i1-i12-i2 such that i1-i12 and i12-i2 done but i1-i2 is not
    [disc,i12s] = dijkstra(GD,i1); done1=done(i1,:);
    i2s = find(ismember(i12s,find(done1)) & ~done1);
    for i2 = i2s
      % get initial snake by merging two shorter snakes, optimize length
      i12=i12s(i2); chi0=[chis{i1,i12}(:,1:end-1) chis{i12,i2}];
      prmSnk.errCase=1; chi=lsmlOptimizeChi(chi0,manifold,prmSnk);
      gd = lengthSnake(chi);
      GD(i1,i2)=gd; GD(i2,i1)=gd; done(i1,i2)=1; done(i2,i1)=1;
      chis{i1,i2}=chi; chis{i2,i1}=fliplr(chi); isDiff = 1;
      tocStatus(ticid,mean(done(:)));
    end
  end
end
GD(done==0)=NaN;
end

function gd = lengthSnake( chi )
gd = sum(sqrt(sum(diff(chi,1,2).^2,1)));
end

function [gd chi] = lsmlGeodesic1( X1,X2,manifold,prmSnk,epsSnk )
% compute geodesic distance between 2 nearby points
chi = createSnake( X1, X2, manifold, epsSnk );
prmSnk.errCase=2; chi=lsmlOptimizeChi( chi,manifold,prmSnk );
prmSnk.errCase=1; chi=lsmlOptimizeChi( chi,manifold,prmSnk );
gd = lengthSnake(chi);
end

function chi = createSnake( chiL, chiR, manifold, epsSnk  )
% create initial snake by interpolating between points chiL and chiR
gapPrev=inf;
for j=1:100
  chiLe=chiL(:,end); chiRe=chiR(:,end);
  chiDiff=chiRe-chiLe; gapCurr=norm(chiDiff);
  if( gapCurr<epsSnk*1.5 );
    % if gap small we're done
    break;
  elseif( gapCurr>.99*gapPrev )
    % if gap did not decrease, split into halves and call recursively
    chiM=(chiLe+chiRe)/2;
    chiLe=createSnake( chiLe, chiM, manifold, epsSnk );
    chiRe=createSnake( chiRe, chiM, manifold, epsSnk );
    chiL=[chiL chiLe(:,2:end-1)]; chiR=[chiR chiRe(:,2:end)]; %#ok<AGROW>
    break;
  else
    % otherwise grow either left or right
    gapPrev=gapCurr; lft = mod(j,2)==1;
    if(lft), chiNxt=chiLe; else chiDiff=-chiDiff; chiNxt=chiRe; end
    for step = 1:10
      H = lsmlComputeH( chiNxt, manifold, true );
      chiDiff = H*H'*chiDiff;
      chiDiff = chiDiff/norm(chiDiff);
      chiNxt = chiNxt + chiDiff*epsSnk/10;
    end
    if(lft), chiL=[chiL chiNxt]; else chiR=[chiR chiNxt]; end %#ok<AGROW>
  end
end
chi = [chiL fliplr(chiR)];
end
