function [GD Y D] = lsmlIsomap( neigh, d )
% Simple implementation of the Isomap algorithm.
%
% Implements the Isomap algorithm described in:
%  A Global Geometric Framework for Nonlinear Dimensionality Reduction
%  Joshua B. Tenenbaum, Vin de Silva, and John C. Langford
%  Science 22 December 2000: 2319-2323.
% Isomap code is also available from the original authors at:
%  http://isomap.stanford.edu/
% The Isomap code here is simpler (although not as optimized) and is
% included for convenience. This function is named lsmlIsomap to avoid any
% naming conflicts with authors' original code.
%
% Not recommended for n>1000.  n=400 takes approximately 1s to run, and
% running time is roughly quadratic.
%
% USAGE
%  [GD Y D] = isomap( neigh, d )
%
% INPUT
%  neigh  - connected neighborhood graph [see computeNeighbor]
%  d      - embedding dimension
%
% OUTPUT
%  GD     - [n x n] distances computed using shortest path
%  Y      - [d x n] coordinates in embedding, using cmdscale
%  D      - [n x n] distances between each point in Y
%
% EXAMPLE
%  data = generateToydata('type',5,'nSet',1,'nPoint',500,'show',0);
%  neigh = computeNeighbor( data.X3,'k',10,'show',0);
%  [GD,Y,D] = lsmlIsomap( neigh, 2 );
%  visualizeManifold(data.X3,1,'S3',data.S3,'Y',Y);
%
% See also COMPUTENEIGHBOR, CMDSCALE, VISUALIZEMANIFOLD
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

n = size(neigh.Dmat,1);

if( ~neigh.conn );
  disp('warning: isomap needs connected graph');
  GD=NaN(n); Y=[]; D=[]; return;
end

GD = dijkstra( neigh.Dmat, 1:n );
Y = cmdscale(GD); Y=Y(:,1:d); Y=Y';
D = pdist2( Y', Y', 'euclidean' );
