function [ Chi gr ] = lsmlOptimizeChi( X, manifold, varargin )
% Helper: perform gradient descent on X for different error functions.
%
% This function optimizes the \Chis according to value of errCase:
%  0: Equations 9 and 10 in ICML07 - denoise
%  1: Equation 11 in ICML07 - minimize snake length
%  2: Equation 10 in ICML07 - make snake stick to the manifold
%  3: only used for image case (unreleased, please ignore)
%
% USAGE
%   [Chi gr] = lsmlOptimizeChi( X, manifold, varargin )
%
% INPUT
%   X           - [ D x n ]  data
%   manifold    - struct returned by lsmlOptimizeTh
%   varargin    - prm struct or name/value list w following fields:
%       .errCase   - 0-denoise; 1-snakes (length); 2-snakes (stick);
%                    3-stick and length for video sequence
%       .errDel    - [1e-3] stopping criteria <-important
%       .nItr      - [500] max number of iterations
%       .show      - [0] optionally display progress in figure(show)
%       .lambdaN   - {only errCase==0} [see lsmlDenoise]
%       .N         - {only errCase==0} neighbors list
%
% OUTPUT
%   Chi        - [D x n]  optimized points
%   gr         - [D x n]  gradient of specified function at each point
%
% EXAMPLE
%
% See also LSMLOPTIMIZETH, LSMLDENOISE, LSMLGEODESICDIST
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

[ errCase errDel nItr show lambdaN N ] = getPrmDflt( varargin, ...
  {'errCase','REQ', 'errDel',[], 'nItr',[], 'show',0,...
  'lambdaN',0, 'N',[]} );
if(isempty(errDel)), errDel=1e-3; end
if(isempty(nItr)), nItr=500; end
D=manifold.D; Chi=X;  n=size(Chi,2);

%%% indexing / neighbors
Nup=cell(1,n);  IJ=zeros(n,n);  m=0;
if( errCase==0 )
  for i1=1:n; Nup{i1}=N{i1}(N{i1}>i1); end
else
  for i1=1:n-1; Nup{i1}=i1+1; end
end
for i1=1:n; for j1=Nup{i1}; m=m+1; IJ(i1,j1)=m; end; end
IJ=IJ+IJ';

%%% main loop -- gradient descent
if( show ); figure(show); clf; tprev=clock; end
ChiAv=zeros(D,m);  BB=zeros(D,m);  ChiDiff=[]; ChiMinusX=[];
eyeD=eye(D);  gamma=0.1;
precomputations();  [errs,errsG,errsX]=getErr;
for itr = 1:nItr
  gr = getGradient; if nargout==2; return; end
  ChiOld = Chi;
  
  %%% move down gradient by some small amount
  for attempt=1:50
    Chi = ChiOld - gamma*gr;
    precomputations();  [err,errG,errX]=getErr;
    if( err<=errs(end) )
      %if(mod(itr,10)==1); gamma=gamma*5; end;
      if(rand()>0.9); gamma=gamma*5; end
      break;
    else
      gamma = gamma/5;
      if gamma<1e-10; break; end
    end
  end
  errs(end+1)=err; errsG(end+1)=errG; errsX(end+1)=errX; %#ok<AGROW>
  
  %%% update display
  if( show && (etime(clock,tprev)>1 || itr==nItr))
    tprev=clock;
    hold on; plot(errs,'r');
    if(errCase==0); plot(errsX,'b'); plot(errsG,'g'); end
    tstr = ['err (r)=' num2str(errs(end))];
    if(errCase==0)
      tstr=[tstr '   errX (b)=' num2str(errsX(end))];  %#ok<AGROW>
      tstr=[tstr '   errG (g)=' num2str(errsG(end))]; %#ok<AGROW>
    end
    tstr = [tstr '   nItr=' int2str(itr)]; %#ok<AGROW>
    xlabel('iteration'); ylabel('error');
    title(tstr);  drawnow;  hold off;
  end
  
  %%% check for small change in error (stopping criteria)
  if( (1-errs(end)/(errs(end-1)+eps))<errDel );
    break;
  end
end

  function precomputations()
    if errCase==1 || errCase==3
      ChiDiff = diff( Chi, 1, 2 );
    end
    if errCase==0 || errCase==2 || errCase==3
      ChiMinusX=Chi-X;
    end
    if errCase==0 || errCase==2
      for i=1:n;
        for j=Nup{i};
          ChiAv(:,IJ(i,j))=(Chi(:,i)+Chi(:,j))/2;
        end
      end
      Hav = lsmlComputeH( ChiAv, manifold, 1 );
      for i=1:n
        for j=Nup{i}; ij=IJ(i,j);
          BB(:,ij) = (eyeD-Hav(:,:,ij)*Hav(:,:,ij)')*(Chi(:,i)-Chi(:,j));
        end
      end
    end
  end

  function gr = getGradient()
    H = lsmlComputeH( Chi, manifold, true );
    gr = zeros(D,n);
    switch errCase
      case {1 3} % minimize length of snake (parallel gr only)
        gr(:,2:end-1)=-2*diff(ChiDiff,1,2);
        if errCase==3; gr=gr+2*lambdaN*ChiMinusX; end
        for i=2:n-1; gr(:,i)=H(:,:,i)*H(:,:,i)'*gr(:, i); end
      case {0 2} % stick to manifold (orthogonal gr only)
        for i=1:n
          for j=Nup{i}
            gr(:,i) = gr(:,i) + 4*BB(:,IJ(i,j));
            gr(:,j) = gr(:,j) - 4*BB(:,IJ(i,j));
          end
        end
        if( errCase==0 ); gr=gr+2*lambdaN*ChiMinusX; end
        if( errCase==2 ); gr(:,[1 end])=0; end
        % project onto the orthogonal to the manifold
        t=gr; for i=1:n, gr(:,i)=H(:,:,i)*H(:,:,i)'*gr(:,i); end, gr=t-gr;
    end
  end

  function [err,errG,errX] = getErr()
    errG = 0; errX = 0;
    switch errCase
      case 0 % de-noising
        errG = 2*norm(BB,'fro')^2;
        errX = lambdaN * norm(ChiMinusX,'fro')^2;
        err = errG + errX;
      case 1 % minimize length of snake
        err = norm(ChiDiff,'fro')^2;
      case 2 % stick snake to manifold
        err = 2*norm(BB,'fro')^2;
      case 3 % stick to the points and minimize the overall length
        err = norm(ChiDiff,'fro')^2 + lambdaN*norm(ChiMinusX,'fro')^2;
    end
  end
end
