function manifold = lsmlOptimizeTh( X, Nmat, varargin )
% Compute manifold warping function.
%
% This function simply otpimizes Equation (5) in ICML07 with respect to the
% \Theta's and the \epsilons.  A closed form of each can be computed, while
% assuming the other is fixed. This gives pretty high convergence rates but
% can suffer from bad initialization (hence the random restarts) and
% ping-pong effect in optimization.
%
% USAGE
%   manifold = lsmlOptimizeTh( X, Nmat, varargin )
%
% INPUT
%   X or X3   - data matrix: [D x n] or [nSet x nPoint x D]
%   Nmat      - [n x n] neighborhood matrix [see computeNeighbor]
%   varargin  - prm struct or name/value list w following fields:
%     [parameters for rbfs]
%       .rbfFlag    - [1]   flag for using rbfs
%       .rbfScale   - [2]   [see rbfComputeBasis]
%       .rbfCluster - [1]   [see rbfComputeBasis]
%       .rbfK       - []    [see rbfComputeBasis]
%     [parameters for coordinate descent]
%       .d          - [REQ]  manifold dimension
%       .nRestart   - [25]   number of restarts (phases)
%       .nItr       - [50]   number of iterations per phase
%       .lam        - [1e-3] regularization term (see Equation 4)
%       .thInit     - [rand] initial value for th
%       .nSamples   - [1000] num to subsample for training
%       .show       - [0]    optionally display progress in figure(show)
%       .percTest   - [20]   percentage of testing data to consider
%
% OUTPUT
%   manifold
%     [dimensions]
%       .D          - original dimension
%       .f          - number of features per point
%       .d          - manifold dimension
%     [rbf]
%       .rbfFlag    - 1 if radial basis functions are being used
%       .rbfBasis   - basis functions ([] if rbfFlag==0)
%     [learned warp]
%       .pw         - parameters used to compute warping
%       .th         - [d x f x D] computed warping function
%       .thErr      - error of computed warping function
%       .thErrF     - non-regularized error of computed warping function
%       .errTest    - test error
%
% EXAMPLE
%
% See Also RBFCOMPUTEBASIS, COMPUTENEIGHBOR, VISUALIZEMANIFOLD
%
% LSML Toolbox     Version 1.1
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

% get parameters (set defaults)
dfs = {'rbfFlag',1, 'rbfScale',2, 'rbfK',[], 'rbfCluster',1, ...
  'd','REQ', 'nRestart',25, 'nItr',50, ...
  'lam',1e-3, 'thInit',[], 'nSamples',1000, 'show',0, 'percTest',20 };
[rbfFlag rbfScale rbfK rbfCluster d nRestart nItr lam thInit  ...
  nSamples show percTest] = getPrmDflt( varargin, dfs );
pw = getPrmDflt( varargin, dfs );

% get training data
if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],size(X,3))'; end
nTrain=ceil(size(Nmat,1)*(1-percTest/100)); D=size(X,1);
[Xbar Del pairs]=prepareData(1:nTrain,Nmat,X,nSamples);
n=size(Xbar,2); lamE=lam; lamTh=lam*n/size(pairs,1);

% run coordinate descent nRestart times, followed by single longer run
if( ~isempty(thInit) ), phase=1; bTh=thInit; bRbfBasis=[]; else
  if(show); figure(show); clf; end
  for phase=1:nRestart
    [th,rbfBasis,err] = computeTh1([],[]);
    if(phase==1 || err<bErr), bErr=err; bTh=th; bRbfBasis=rbfBasis; end
  end
end
if(show); figure(show); clf; end
nItr=nItr*10; [th,rbfBasis,err,errF]=computeTh1(bTh,bRbfBasis);

% store results
manifold = struct( 'D',D, 'f',f, 'd',d, 'rbfFlag',rbfFlag, ...
  'rbfBasis',rbfBasis, 'pw',pw, 'th',th, 'thErr',err, 'thErrF',errF );

% Compute the test error
if( nTrain < size(Nmat,1) )
  [XbarTst DelTst]=prepareData(nTrain+1:size(Nmat,1),Nmat,X,nSamples);
  H=lsmlComputeH(XbarTst,manifold); err=0;
  for i0=1:size(XbarTst,2)
    err=err + norm( H(:,:,i0)*(H(:,:,i0)\DelTst(:,i0))-DelTst(:,i0) )^2;
  end
  manifold.errTest = err/size(XbarTst,2);
end

  function [th,rbfBasis,err,errF] = computeTh1( th, rbfBasis )
    % run coordinate descent once
    
    % get rbfBasis if not provided
    if( ~rbfFlag ), rbfBasis=[]; f=D; else
      if( isempty(rbfBasis));
        if( size(X,2)<nSamples ), X1=X; else
          X1=X(:,randsample(size(X,2),nSamples));
        end
        rbfBasis = rbfComputeBasis(X1',rbfK,rbfCluster,rbfScale,0);
      end
      f = rbfBasis.k;
    end
    
    % randomly initialize th if not provided
    if( isempty(th) ); th=rand(d,f,D)-0.5; end
    
    % pre-computations
    if(rbfFlag); F=rbfComputeFtrs(Xbar',rbfBasis)'; else F=Xbar; end
    delF = (F(:,pairs(:,1)) - F(:,pairs(:,2)))';
    delFF = lamTh * kron( eye(d), delF'*delF );
    tprev=clock; eps=zeros(d,n); H=zeros(D,d,n);
    krondf=kron(1:d,ones(1,f)); Fd=repmat(F,[d 1]);
    errs=zeros(1,nItr); errsF=zeros(1,nItr);
    
    % main loop
    for itr=1:nItr
      computeEfromTh;
      computeThFromE;
      if( show )
        [errs(itr),errsF(itr)] = getErr;
        if((etime(clock,tprev)>1 || itr==nItr))
          tprev=clock; hold on; plot(errs,'r'); hold on; plot(errsF,'b');
          title( sprintf('err=%f   errF=%f   nItr=%i   phase=%i', ...
            errs(itr), errsF(itr), itr, phase ) );
          xlabel('iteration'); ylabel('error'); drawnow; hold off;
        end
      end
    end
    if(show), err=errs(end); errF=errsF(end); else [err errF]=getErr; end
    
    function computeThFromE
      % compute th given E
      At = eps(krondf,:).*Fd;
      th = permute( reshape((At*At'+delFF)\(At*Del'),[f d D]), [2 1 3]);
    end
    
    function computeEfromThOld %#ok<DEFNU>
      % compute E given th - slow but easy to understand
      for k=1:D; H(k,:,:)=th(:,:,k)*F; end; eyeE=lamE*eye(d);
      for i=1:n; Hi=H(:,:,i); eps(:,i)=(Hi'*Hi+eyeE)\(Hi'*Del(:,i)); end;
    end
    
    function computeEfromTh
      % compute E given th - without loops (fast)
      H=reshape(permute(th,[1 3 2]),d*D,f)*F;
      H=permute(reshape(H,d,D,n),[2 1 3]);
      AA1=reshape(permute(H,[2,1,3]),[d D 1 n]); AA2=reshape(H,[1 D d n]);
      AA=reshape(sum(bsxfun(@times,AA1,AA2),2),[d d n]); % Hi'*Hi
      for k=1:d; AA(k,k,:)=AA(k,k,:)+lamE; end % Hi'*Hi+eyeE
      BB = reshape(sum(bsxfun(@times, permute(H,[2,1,3]), ...
        reshape(Del,[1 D n ])),2),d,n);  % Hi'*Del(:,i)
      step=ceil(10000/d/d); s=0; % AA(:,:,i)\BB(:,i) (in chunks)
      while(1), if(s==n), break; end; e=min(s+step,n);
        eps(s*d+1:e*d)=blkdiag2(AA(:,:,s+1:e))\BB(s*d+1:e*d)'; s=e;
      end
    end
    
    function [err,errF] = getErr
      % compute error
      errE = norm( eps, 'fro' )^2;
      errTh= norm( delF*reshape(permute(th,[2 1 3]),f,[]), 'fro' )^2;
      errF = norm( squeeze( multiprod(H,eps,[1 2],1) ) - Del, 'fro' )^2;
      err = errF + lamE*errE + lamTh*errTh;
      err=err/n*10^4; errF=errF/n*10^4;
    end
  end
end

function B = blkdiag2( A )
% creates sparse block diagonal matrix using the k blocks in [m x n x k] A
[m,n,k]=size(A); persistent m0 n0 k0 rs cs;
if( isempty(m0) || m~=m0 || n~=n0 || k~=k0 )
  ds=(1:m)'; rs=reshape(1:m*k,m,k); rs=rs(ds(:,ones(1,n)),:); rs=rs(:);
  cs=1:n*k; cs=cs(ones(m,1),:); cs=cs(:); m0=m; n0=n; k0=k;
end
B=sparse(rs,cs,A(:),m*k,n*k);
end

function [Xbar Del pairs] = prepareData( range, Nmat, X, nSamples )
% get all the possible neighboring couples, optionally subsample
[y,x] = find(Nmat(range,range));  NPair = [ x(y>x), y(y>x) ];
if( nSamples<size(NPair,1) )
  NPair = NPair( randsample(size(NPair,1), nSamples), : );
end
NPair=[range(NPair(:,1))',range(NPair(:,2))'];

% get the corresponding pair indices
uNPair=unique(NPair(:)); pairs = cell(length(uNPair),1);
for i=1:length(uNPair)
  [ NPairi, disc ] = find(NPair==uNPair(i)); %#ok<NASGU>
  [ temp1, temp2 ] = meshgrid( NPairi ); temp = temp1(:)<temp2(:);
  pairs{i} = [ temp1(temp), temp2(temp) ];
end
pairs = cell2mat( pairs );

Xbar = 0.5*( X(:,NPair(:,1)) + X(:,NPair(:,2)) );
Del = X(:,NPair(:,2)) - X(:,NPair(:,1));
end
