function manifold = lsmlOptimizeThUnc( X, N, varargin )
% Compute manifold warping function (deprecated: use lsmlOptimizeTh)
%
% This is similar to lsmlOptimizeTh except that is uses the uncentered
% derivative assumption (used in NIPS06), which performs worse than the
% centered/symmetric approximation as shown in ICML07. This function should
% be considered DEPRECATED except for comparisons with centered version. It
% may also not have the latest functionality of lsmlOptimizeTh.
%
% USAGE
%   manifold = lsmlOptimizeThUnc( X, N, varargin )
%
% INPUT
%   X or X3   - data matrix: [D x n] or [nSet x nPoint x D]
%   N         - [nxNi] cell array of neighbors [see computeNeighbor]
%   varargin  - prm struct or name/value list w following fields:
%     [parameters for rbfs]
%       .rbfFlag    - [1]   flag for using rbfs
%       .rbfScale   - [2]   [see rbfComputeBasis]
%       .rbfCluster - [1]   [see rbfComputeBasis]
%       .rbfK       - []    [see rbfComputeBasis]
%     [parameters for coordinate descent]
%       .d          - [REQ]  manifold dimension
%       .nRestart   - [25]   number of restarts (phases)
%       .nItr       - [50]   number of iterations per phase
%       .lam        - [1e-3] regularization term
%       .thInit     - [rand] initial value for th
%       .show       - [0]    optionally display progress in figure(show)
%       .p          - [2*d]  num principle comp
%
% OUTPUT
%   manifold
%     [dimensions]
%       .D          - original dimension
%       .f          - number of features per point
%       .d          - manifold dimension
%     [rbf]
%       .rbfFlag    - 1 if radial basis functions are being used
%       .rbfBasis   - basis functions ([] if rbfFlag==0)
%     [learned warp]
%       .pw         - parameters used to compute warping
%       .th         - [d x f x D] computed warping function
%       .thErr      - error of computed warping function
%
% EXAMPLE
%
% See Also LSMLOPTIMIZETH
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

% get parameters (set defaults)
dfs = {'rbfFlag',1, 'rbfScale',2, 'rbfK',[], 'rbfCluster',1, ...
  'd','REQ', 'nRestart',25, 'nItr',50, ...
  'lam',1e-3, 'thInit',[], 'show',0, 'p',[] };
[rbfFlag rbfScale rbfK rbfCluster d nRestart nItr lam thInit  ...
  show p] = getPrmDflt( varargin, dfs );
pw = getPrmDflt( varargin, dfs );
if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],size(X,3))'; end;
[D n] = size(X);  if(isempty(p)); p=min(D,2*d); end;

% precomputations
US = zeros( D, p, n );
for i1=1:n
  cnti=length(N{i1}); if( cnti==0 ); continue; end;
  DELi = X(:,N{i1})-repmat(X(:,i1),[1 cnti]);
  [Ui,Si] = svd( DELi,'econ');  USi = Ui * Si;
  pi = min(p,size(USi,2)); US(:,1:pi,i1) = USi(:,1:pi);
end;
maxerr=sum(US(:).^2); minerr=US(:,1:d,:);
minerr=1-(sum(minerr(:).^2)/maxerr);
USflat = reshape(US,D,[])';  %p * nxD
eyeE  = lam * eye(d);

% run coordinate descent nRestart times
if(show); figure(show); clf; end;
if( ~isempty( thInit) )
  phase = 1; nItr=nItr*10;
  [th,rbfBasis,err] = computeTh1( thInit );
else
  bestErr=inf;
  for phase=1:nRestart
    [th,rbfBasis,err] = computeTh1;
    if( err < bestErr )
      bestErr=err; bestTh=th; bestBasis=rbfBasis;
    end
  end
  if(show); figure(show); clf; end
  nItr = nItr*10;
  [th,rbfBasis,err] = computeTh1( bestTh, bestBasis );
end

% store results
manifold = struct( 'D',D, 'f',f, 'd',d, 'rbfFlag',rbfFlag, ...
  'rbfBasis',rbfBasis, 'pw',pw, 'th',th, 'thErr',err );

  function [th,rbfBasis,err] = computeTh1( th, rbfBasis )
    % run coordinate descent once
    
    % get rbfBasis functions if not provided
    if( nargin<2 || isempty(rbfBasis));
      if( rbfFlag )
        rbfBasis = rbfComputeBasis(X',rbfK,rbfCluster,rbfScale,0);
        f = rbfBasis.k;
      else
        rbfBasis=[]; f=D;
      end;
    end;
    
    % randomly initialize th if not provided
    if( nargin<1 || isempty(th));
      th=rand(d,f,D)-0.5;
    end;
    
    % pre-computations
    if(rbfFlag); F=rbfComputeFtrs(X',rbfBasis)'; else F=X; end;
    delF=zeros(n*n,f);  cnt=0;
    for i2=1:n
      for j=N{i2}
        cnt=cnt+1;  delF(cnt,:)=F(:,j)'-F(:,i2)';
      end;
    end;
    delF  = delF(1:cnt,:);
    delFF = lam * kron( eye(d), delF'*delF );
    
    % pre-computations for fastkron
    [ia,ib]=meshgrid(1:p,1:1);  [ja,jb]=meshgrid(1:d,1:f);
    KFT=zeros(p,d*f,n); for i2=1:n; KFT(:,:,i2)=F(jb,i2*ib)'; end;
    
    % main loop
    if( nargin<1 || isempty(th)); th=rand(d,f,D)-0.5; end;
    tprev=clock;  E=zeros(d,p,n);  H=zeros(D,d,n);
    computeEfromTh;  [errs,errsF]=getErr;
    for itr=1:nItr
      computeThFromE;
      computeEfromTh;
      [errs(end+1),errsF(end+1)] = getErr; %#ok<AGROW>
      if( show && (etime(clock,tprev)>1 || itr==nItr) )
        tprev=clock;
        hold on; plot(errs,'r'); hold on; plot(errsF,'b');
        xlabel('iteration'); ylabel('error');
        tstr = ['err=' num2str(errs(end)) ...
          '   errF=' num2str(errsF(end)) ...
          '   nItr=' int2str(itr) ...
          '   phase=' int2str(phase)];
        title( tstr );  drawnow;  hold off;
      end
    end
    err=errs(end);
    
    function K = fastkronEF( i )
      % fast inline versin of "kron(E(:,:,i)', F(:,i)')"
      K = E(ja,ia,i)'.*KFT(:,:,i);
    end
    
    function computeThFromE
      % compute th given E
      KRk = zeros( n*p, d*f );
      for k=1:D
        for i=1:n; KRk(((i-1)*p+1):(i*p),:)=fastkronEF(i);  end;
        thk = (KRk'*KRk + delFF) \ (KRk' * USflat(:,k));
        th(:,:,k) = reshape( thk, f, d )';
      end
    end
    
    function computeEfromTh
      % compute E given th
      for k=1:D; H(k,:,:)=th(:,:,k)*F;  end;
      for i=1:n; Hi=H(:,:,i); E(:,:,i)=(Hi'*Hi+eyeE) \ Hi'*US(:,:,i); end;
    end
    
    function [err,errF] = getErr
      % compute error
      err = sum(E(:).^2); errF=0;
      for k=1:D; err=err+sum(sum( (delF*th(:,:,k)').^2 )); end;
      for i=1:n; errF=errF+sum(sum((H(:,:,i)*E(:,:,i)-US(:,:,i)).^2)); end;
      err = (errF + lam*err) / maxerr - minerr;
      errF = errF / maxerr - minerr;
    end
  end
end
