function visualizeManifold( X, show, varargin )
% Manifold visualization functions for data in D=2 or D=3.
%
% What actually gets displayed depends on the contents of prm. For example,
% to display transparent true surface, S3 should be set. This function has
% a wide diversity of uses, see calling functions.
%
% DIMENSIONS
%   nSet   - number of point sets
%   nPoint - number of points / set
%   n      - total number of points (nSet*nPoint)
%   D      - original dimension (2 or 3)
%   d      - manifold dimension (1 or 2)
%
% USAGE
%   visualizeManifold( X, show, prm )
%
% INPUT
%   X or X3  - data matrix: [D x n] or [nSet x nPoint x D]
%   show     - figure to use for display
%   varargin - list of paramaters in quotes alternating with their values
%       'VF'    - [D x d x n] tangent vectors at each point
%       'N'     - list of neighbor connections [see computeNeighbor]
%       'Nmat'  - [nxn] sparse matrix st N(i,j)=1 iff i,j are neighbors
%       'Y'     - [d x n] embedding of data points
%       'Snake' - {nSnake} cell of all the snakes to display. [D x ni]
%       'S3'    - representation of surfaces -- see generateToydata
%       'clfF'  - if [true] clears figure before display
%       'Chi3'  - denoised points (displays original/denoised in red/blue)
%       'fName' - filename to save to
%
% EXAMPLE
%
% See also GENERATETOYDATA COMPUTENEIGHBOR
%
% LSML Toolbox     Version 1.0
% Copyright 2009 Piotr Dollar & Vincent Rabaud
% Contact [pdollar-at-caltech.edu] & [vrabaud-at-cs.ucsd.edu]
% Please email us if you find bugs, or have suggestions or questions!
% Licensed under the Lesser GPL [see lgpl.txt]

%%% parameters
[ fName VF N Nmat S3 clfF Chi3 Y Snake] = getPrmDflt( varargin, ...
  { 'fName',[], 'VF',[], 'N',[], 'Nmat', [], 'S3',[], 'clfF',1,...
  'Chi3',[], 'Y',[], 'Snake', []} );

%%% data
if(ndims(X)==3); X=reshape(permute(X,[2 1 3]),[],size(X,3) )'; end;
[D,n]=size(X);
if(~(D==2 || D==3)); error('D must be 2 or 3'); end;
if(~isempty(Y) && size(Y,2)~=n); error( 'bad Y' ); end;
if(isempty(Y)); C=X(end,:)'; else C=Y(end,:)'; end;
if(~isempty(Chi3)); Chi=reshape(permute(Chi3,[2 1 3]),[],D)'; end;

%%% figure setup
if(clfF); figure(show); clf; end;
if(~isempty(Y)); subplot(1,2,1); end;
colormap hsv; grid off; axis off; hold on;

%%% display surfaces
if( ~isempty(S3) )
  if( D==2 )
    for s=1:length(S3)
      lnh=line( S3(s).X, S3(s).Y );
      set(lnh,'Color',[.7, .7, .7],'LineWidth',2.5);
    end
  else
    for s=1:length(S3)
      S3(s).C = repmat(.25,[size(S3(s).X) 3]);
      surf(S3(s).X,S3(s).Y,S3(s).Z,S3(s).C);
      shading flat;
      alpha(.2);
    end
  end
end

%%% Show vector field VF at points X.
if( ~isempty(VF) )
  if(size(VF,1)~=D || size(VF,3)~=n); error( 'bad VF' ); end
  d = size(VF,2);
  for i=1:d
    VFi=squeeze(VF(:,i,:));
    for j=1:size(VFi,2); VFi(:,i)=VFi(:,i)/norm(VFi(:,i)); end
    if( D==2 )
      h=quiver( X(1,:), X(2,:), VFi(1,:), VFi(2,:), .25 );
    else
      h=quiver3( X(1,:), X(2,:), X(3,:), VFi(1,:), VFi(2,:), VFi(3,:) );
    end
    set(h,'LineWidth',2 );
  end
end

%%% display each snake
if( ~isempty(Snake) )
  for i=1:length( Snake );
    ni = size(Snake{i},2);
    Ni=cell(ni,1); for j=2:ni; Ni{j}=[j-1,j+1]; end;
    if(ni>1); Ni{1}=2; Ni{ni}=ni-1; end;
    visualizeManifold(Snake{i},show,'N',Ni,'clfF',0);
  end
end

%%% display neighbor connections
if( ~isempty(Nmat) )
  N=cell(1,n);
  for i=1:n
    N{i}=find(Nmat(i,:));
  end
end

if( ~isempty(N) )
  for i=1:n
    for j=N{i}%(N{i}>i);
      if( D==2 )
        h=line([X(1,i); X(1,j)],[X(2,i); X(2,j)]);
      else
        h=line([X(1,i); X(1,j)],[X(2,i); X(2,j)],[X(3,i); X(3,j)]);
      end
      set(h,'LineWidth',2,'Color','k');
    end
  end
end

%%% show original points [and possibly noisy points]
if(isempty(Chi3))
  if( D==2 )
    scatter( X(1,:), X(2,:), 40, C, 'filled' );
  else
    scatter3( X(1,:), X(2,:), X(3,:), 60, C, 'filled' );
  end
else
  if( D==2 )
    scatter( X(1,:), X(2,:), 40, 'r', 's', 'filled' );
    scatter( Chi(1,:), Chi(2,:), 40, 'b', 'filled' );
  else
    scatter3( X(1,:), X(2,:), X(3,:), 60, 'r', 's', 'filled' );
    scatter3( Chi(1,:), Chi(2,:), Chi(3,:), 60, 'b', 'filled' );
  end
end

%%% set axis and view
if( D==3 ); view([12 -20 3]); end
if(clfF); niceAxis(); hold off; end
set( gcf, 'Color', 'w' );

%%% display embedding if given
if(~isempty(Y));
  d = size(Y,1);
  if( d==2 )
    subplot(1,2,2);
    scatter( Y(1,:), Y(2,:), 40, C, 'filled' );
    colormap hsv; grid off; axis off; axis equal; axis tight;
  end
end

%%% optionally save
if(~isempty(fName));
  savefig( fName, 1, 'pdf', '-r400' );
end

end

function niceAxis()
% plot nice axis bars (modified from LLE examples)
axis equal; a=axis; mins=a(1:2:end); maxs=a(2:2:end);
r=maxs-mins; mins=mins-r/10; maxs=maxs+r/10;
D=length(mins); LC=cell(1,D);
for i=1:D, LC{i}=repmat(mins(i),2,3); LC{i}(2,i)=maxs(i); end
line( LC{:}, 'Color',[.6,.6,.6], 'LineWidth',2.5,...
  'LineStyle','-', 'Clipping','off' );
axis tight;
end
